import asyncio
import logging
import multiprocessing
import socket
from threading import Thread

import rich
from rich.panel import Panel

from decentnet.consensus.dev_constants import RUN_IN_DEBUG
from decentnet.consensus.difficulty_ports_mapping import PORT_DIFFICULTY_CONFIG
from decentnet.consensus.net_constants import BLOCKED_IPV4, BLOCKED_IPV6
from decentnet.modules.comm.relay import Relay
from decentnet.modules.logger.log import setup_logger
from decentnet.modules.tasks_base.consumer import Consumer
from decentnet.modules.tasks_base.r2r_comm import R2RComm
from decentnet.modules.tcp.db_functions import remove_alive_beam_from_db_w_pub_key
from decentnet.modules.timer.timer import Timer

logger = logging.getLogger(__name__)

setup_logger(RUN_IN_DEBUG, logger)


class TCPServer:
    def __init__(self, host, port):
        self.host = host
        self.port = port

        if ":" not in host:
            self.server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.server.bind((self.host, self.port))
        else:
            self.server = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
            # Allow dual-stack (IPv4 + IPv6) connections
            self.server.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_V6ONLY, 0)
            self.server.bind((self.host, self.port, 0, 0))

        self.clients = []

    @staticmethod
    def create_socket(host, port):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.connect((host, port))
        return sock

    @staticmethod
    def handle_client(client_socket,
                      beam_pipe_comm: dict, beam_message_queue: dict, logs_dict: dict):
        """
        Handle a client connection

        ! This function is run by separate process
        :param logs_dict: Logs dict of pipes
        :param beam_message_queue: Beam multiprocess message queue
        :param client_socket:
        :param beam_pipe_comm:
        :return:
        """
        relay = None
        try:
            relay = Relay(client_socket, beam_pipe_comm, beam_message_queue, logs_dict)
        except ConnectionError:
            logger.debug("Beacon disconnected before providing genesis block.")
            return

        try:
            sock_name = client_socket.getsockname()
        except OSError as e:
            logger.error(e)
            logger.info("Socket was closed, unable to continue.")
            return

        Thread(target=lambda: R2RComm(relay), name=f"R2R Communication {relay.target_key}",
               daemon=True).start()
        Thread(target=lambda: Consumer(relay, beam_message_queue),
               name=f"Consumer Runner {client_socket}", daemon=True).start()
        t = Timer()
        while True:
            try:
                if not relay.do_relaying(t):
                    TCPServer.graceful_shutdown(beam_message_queue, beam_pipe_comm, relay, sock_name)
                    break
            except (ConnectionError, ConnectionResetError, ConnectionAbortedError,
                    ConnectionRefusedError):
                TCPServer.graceful_shutdown(beam_message_queue, beam_pipe_comm, relay, sock_name)
                break

    @staticmethod
    def graceful_shutdown(beam_message_queue, beam_pipe_comm, relay, sock_name):
        logger.info(
            f"Disconnected {sock_name[0]}:{relay.local_port}")  # TODO: Delete from connected
        asyncio.run(relay.disconnect_beacon(relay.local_port, relay.beam_pub_key))
        asyncio.run(remove_alive_beam_from_db_w_pub_key(relay.beam_pub_key))

        if relay.beam_pub_key:
            beam_pipe_comm[relay.beam_pub_key][0].close()
            beam_message_queue[relay.beam_pub_key][0].close()
            beam_pipe_comm[relay.beam_pub_key][1].close()
            beam_message_queue[relay.beam_pub_key][1].close()
            beam_message_queue.pop(relay.beam_pub_key, None)
            beam_pipe_comm.pop(relay.beam_pub_key, None)

    def run(self, logs_dict: dict):
        diff_setting = PORT_DIFFICULTY_CONFIG[self.port]
        self.server.listen(diff_setting.max_hosts)
        logger.info(f"[*] Listening on {self.host}:{self.port}")
        welcome_string = (f"Started DecentMesh... OK\n"
                          f"Listening on {self.host}:{self.port}\n"
                          f"Listening for maximum hosts {diff_setting.max_hosts}\n"
                          f"Seed difficulty: {diff_setting.seed_difficulty}\n"
                          f"Low difficulty: {diff_setting.low_diff}")
        rich.print(Panel(welcome_string, title="DecentMesh Status"))
        manager = multiprocessing.Manager()
        beam_pipe_comm = manager.dict()
        beams_message_queue = manager.dict()

        while True:
            client_socket, client_address = self.server.accept()

            if client_address in (BLOCKED_IPV4 + BLOCKED_IPV6):
                logger.debug(f"Blocked {client_address}, found in blacklist.")
                client_socket.close()
                continue

            logger.info(
                f"[*] Accepted connection from {client_address[0]}:{client_address[1]}")
            client_process = multiprocessing.Process(target=TCPServer.handle_client,
                                                     name=f"Handling client {client_address}",
                                                     args=(
                                                         client_socket,
                                                         beam_pipe_comm,
                                                         beams_message_queue,
                                                         logs_dict), daemon=True)
            client_process.start()

    def close(self):
        for client_socket in self.clients:
            client_socket.close()
        self.server.close()



import logging

import ecies
import pyximport
from ecdsa import SigningKey, VerifyingKey, BadSignatureError, NIST256p

from decentnet.consensus.dev_constants import RUN_IN_DEBUG
from decentnet.modules.logger.log import setup_logger

logger = logging.getLogger(__name__)

setup_logger(RUN_IN_DEBUG, logger)

pyximport.install()
try:
    from decentnet.modules.convert.byte_to_base64_fast import *  # noqa
except ImportError as ex:
    logger.debug(f"Operating in slow mode because of {ex}")
    logger.warning("Base64 convert module is operating in slow mode")
    from decentnet.modules.convert.byte_to_base64_slow import *


class AsymCrypt:

    @staticmethod
    def generate_key_pair_encryption() -> [bytes, bytes]:
        """Generate a key pair for ECIES."""
        private_key = ecies.utils.generate_key()
        sk_bytes = private_key.secret  # bytes
        pk_bytes = private_key.public_key.format(True)  # bytes
        return sk_bytes, pk_bytes

    @staticmethod
    def generate_key_pair_signing(curve=NIST256p) -> [SigningKey, VerifyingKey]:
        """Generate a key pair for ECDSA."""
        # Generate a private key
        private_key = SigningKey.generate(curve=curve)

        # Derive the public key from the private key
        public_key = private_key.verifying_key

        return private_key, public_key

    @staticmethod
    def check_encryption_key(public_key: bytes):
        try:
            ecies.encrypt(public_key, b"test")
            return True
        except TypeError:
            return False

    @staticmethod
    def encrypt_message(public_key: bytes, data: bytes) -> bytes:
        """Encrypts data using ECIES."""
        encrypted_data = ecies.encrypt(public_key, data)
        return encrypted_data

    @staticmethod
    def decrypt_message(private_key: bytes, encrypted_data: bytes) -> bytes:
        """Decrypts data using ECIES."""
        decrypted_data = ecies.decrypt(private_key, encrypted_data)
        return decrypted_data

    @staticmethod
    def sign_message(private_key: SigningKey, data: bytes) -> bytes:
        """Signs data using ECDSA."""
        if type(private_key) is not SigningKey:
            raise Exception("Not a Signing Key")
        signature = private_key.sign(data)
        return signature

    @staticmethod
    def verify_signature(*, public_key: VerifyingKey, signature: bytes,
                         data: bytes) -> bool:
        """Verifies the signature of data using ECDSA."""
        if type(public_key) is not VerifyingKey:
            raise Exception("Not a Verifying Key")

        try:
            return public_key.verify(signature, data)
        except BadSignatureError:
            return False

    @staticmethod
    def key_pair_from_private_key_base64(private_key: str, can_encrypt: bool) -> tuple[
                                                                            SigningKey,
                                                                            VerifyingKey] | \
                                                                                 tuple[
                                                                            bytes, bytes]:
        """
        Returns an Object key pair from base64 str
        :param private_key:
        :param can_encrypt:
        :return: Private Key Object, Public Key object
        """
        if not can_encrypt:
            private_key = SigningKey.from_string(base64_to_original(private_key),
                                                 curve=NIST256p)
            public_key = private_key.verifying_key
            return private_key, public_key
        else:
            pk_bytes = base64_to_original(private_key)
            private_key = ecies.PrivateKey(pk_bytes)
            public_key = private_key.public_key.format(True)
            return private_key.secret, public_key

    @staticmethod
    def public_key_from_base64(public_key: str,
                               can_encrypt: bool) -> VerifyingKey | bytes:
        """
        Returns Object key from base64 str
        :param public_key:
        :param can_encrypt:
        :return: Private Key Object, Public Key object
        """

        if not can_encrypt:
            return AsymCrypt.verifying_key_from_string(public_key)
        else:
            pk_bytes = base64_to_original(public_key)
            return ecies.PublicKey(pk_bytes).format(True)

    @staticmethod
    def encryption_key_to_base64(key: bytes) -> str:
        """
        Converts bytes of public or private key base64
        :param key:
        :return:
        """
        return bytes_to_base64(key)

    @staticmethod
    def verifying_key_to_string(key: VerifyingKey) -> str:
        """
        Converts a verification key into string base64
        :param key:
        :return:
        """
        return bytes_to_base64(key.to_string("compressed"))

    @staticmethod
    def verifying_key_from_string(key: str) -> VerifyingKey:
        return VerifyingKey.from_string(base64_to_original(key),
                                        curve=NIST256p)



import time
import unittest

from decentnet.modules.compression.wrapper import CompressionWrapper


class TestCompressionWrapperPerformance(unittest.TestCase):

    def test_compression_performance(self):
        original_data = b'"activities": ["login", "update_profile", "logout"]}' * 10 ** 4  # 1 MB of repetitive data, adjust as needed

        # Define the ranges for compression levels and block sizes
        compression_levels = range(5, 21)  # Compression levels from 5 to 20
        block_sizes = [64 * 1024, 128 * 1024, 256 * 1024]  # Block sizes: 64KB, 128KB, 256KB

        for level in compression_levels:
            for block_size in block_sizes:
                with self.subTest(compression_level=level, block_size=block_size):
                    # Measure compression time
                    start_time = time.time()
                    compressed_data = CompressionWrapper.compress_lz4(original_data, level)
                    compression_time = time.time() - start_time

                    # Ensure compression was successful
                    self.assertIsNotNone(compressed_data)
                    self.assertGreater(len(compressed_data), 0)

                    # Measure decompression time
                    start_time = time.time()
                    decompressed_data = CompressionWrapper.decompress_lz4(compressed_data)
                    decompression_time = time.time() - start_time

                    # Ensure decompression was successful and data integrity is maintained
                    self.assertEqual(original_data, decompressed_data)

                    # Calculate compression ratio
                    compression_ratio = len(compressed_data) / len(original_data)

                    # Print performance metrics
                    print(f"Level: {level}, Block Size: {block_size // 1024} KB, "
                          f"Compression Time: {compression_time:.6f}s, "
                          f"Decompression Time: {decompression_time:.6f}s, "
                          f"Compression Ratio: {compression_ratio:.4f}")


if __name__ == '__main__':
    unittest.main()



import logging

from decentnet.consensus.dev_constants import R2R_LOG_LEVEL
from decentnet.modules.blockchain.block import Block
from decentnet.modules.logger.log import setup_logger

logger = logging.getLogger(__name__)

setup_logger(R2R_LOG_LEVEL, logger)


class R2RComm:
    def __init__(self, relay):
        self.relay = relay
        logger.debug(f"Current pipes {relay.beam_pipe_comm.keys()}")
        process_uid = relay.beam_pub_key

        while relay.alive:
            self.relay_key = process_uid
            if process_uid in relay.beam_pipe_comm.keys():
                logger.debug(f"Receiving PIPE on {process_uid} for sync")
                data = relay.beam_pipe_comm[process_uid][0].recv()
                logger.debug("Received data from PIPE %s B" % (len(data)))
                block = Block.from_bytes(data)

                if block.index == 0:
                    logger.debug(f"Overwriting blockchain with new genesis R2R")
                    relay.beam.comm_bc.clear()

                relay.beam.comm_bc.difficulty = block.diff
                relay.beam.comm_bc.insert(block)

        logger.debug(f"Closing blockchain sync pipes for  {process_uid}")
        self.close_relay_pipe()

    def close_relay_pipe(self):
        self.relay.beam_pipe_comm[self.relay_key][0].close()
        self.relay.beam_pipe_comm[self.relay_key][1].close()
        self.relay.beam_pipe_comm.pop(self.relay_key)



#cython: language_level=3
import argon2

from decentnet.consensus.blockchain_params import BlockchainParams
from decentnet.consensus.byte_conversion_constants import ENDIAN_TYPE

def int_to_bytes(num: int, buffer=None):
    if num == 0:
        return b'\x00'

        # Calculate the number of bytes needed to represent the integer
    byte_length = (num.bit_length() + 7) // 8

    # If no buffer is provided or it's too small, return the bytes directly
    if buffer is None or len(buffer) < byte_length:
        return num.to_bytes(byte_length, ENDIAN_TYPE)

    # Convert integer to bytes directly into the provided buffer
    num_bytes = num.to_bytes(byte_length, ENDIAN_TYPE)

    buffer[:byte_length] = num_bytes
    return bytes(buffer[:byte_length])

def compute_pow(n_bits, hash_t, nonce):
    _bits = hash_t.diff.hash_len_chars * 8 - n_bits
    # Clone the original hash_t to avoid modifying the original object

    while int.from_bytes(argon2.hash_password_raw(int_to_bytes(hash_t.value_as_int() + nonce),
                                                  BlockchainParams.default_salt, hash_t.diff.t_cost,
                                                  hash_t.diff.m_cost,
                                                  hash_t.diff.p_cost, hash_t.diff.hash_len_chars),
                         ENDIAN_TYPE).bit_length() > _bits:
        nonce += 1

    # Return the correct nonce but do not modify the original hash_t
    return nonce
